/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 Florian Spreckelsen <f.spreckelsen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/entity.h"
#include "caosdb/exceptions.h"
#include "caosdb/data_type.h"                       // for DataType
#include "caosdb/entity/v1alpha1/main.pb.h"         // for Messages
#include "caosdb/protobuf_helper.h"                 // for get_arena
#include "caosdb/value.h"                           // for Value
#include <boost/algorithm/string.hpp>
#include <google/protobuf/arena.h>                  // for Arena
#include <google/protobuf/generated_message_util.h> // for Arena::Create...
#include <new>                                      // for operator new

namespace caosdb::entity {
using caosdb::entity::v1alpha1::IdResponse;
using ProtoParent = caosdb::entity::v1alpha1::Parent;
using ProtoProperty = caosdb::entity::v1alpha1::Property;
using ProtoEntity = caosdb::entity::v1alpha1::Entity;
using ProtoImportance = caosdb::entity::v1alpha1::Importance;
using caosdb::entity::v1alpha1::EntityRole;
using ProtoMessage = caosdb::entity::v1alpha1::Message;
using ProtoFileDescriptor = caosdb::entity::v1alpha1::FileDescriptor;
using caosdb::utility::get_arena;
using google::protobuf::Arena;

Messages::~Messages() = default;

// Forward declarations ///////////////////////////////////////////////////////

template<typename E>
auto FixValueImpl(E* ent) -> void;

// Parent /////////////////////////////////////////////////////////////////////

Parent::Parent() : wrapped(Parent::CreateProtoParent()) {
  // TODO(fspreck) Re-enable once we have decided how to attach
  // messages to parents.
  // errors.wrapped = this->wrapped->mutable_errors();
  // warnings.wrapped = this->wrapped->mutable_warnings();
  // infos.wrapped = this->wrapped->mutable_infos();
}

auto Parent::CreateProtoParent() -> ProtoParent * {
  return Arena::CreateMessage<ProtoParent>(get_arena());
}

auto Parent::SetName(const std::string &name) -> void { this->wrapped->set_name(name); }

auto Parent::SetId(const std::string &id) -> void { this->wrapped->set_id(id); }

[[nodiscard]] auto Parent::GetId() const -> const std::string & { return this->wrapped->id(); }

[[nodiscard]] auto Parent::GetName() const -> const std::string & { return this->wrapped->name(); }

[[nodiscard]] auto Parent::GetDescription() const -> const std::string & {
  return this->wrapped->description();
}

Property::Property() : Property(Property::CreateProtoProperty()) {
  FixValue();
}

auto Property::CreateProtoProperty() -> ProtoProperty * {
  return Arena::CreateMessage<ProtoProperty>(get_arena());
}

[[nodiscard]] auto Property::GetId() const -> const std::string & { return this->wrapped->id(); }

[[nodiscard]] auto Property::GetName() const -> const std::string & {
  return this->wrapped->name();
}

[[nodiscard]] auto Property::GetDescription() const -> const std::string & {
  return this->wrapped->description();
}

[[nodiscard]] auto Property::GetImportance() const -> Importance {
  return static_cast<Importance>(this->wrapped->importance());
}

[[nodiscard]] auto Property::GetValue() const -> const Value & { return this->value; }

[[nodiscard]] auto Property::GetUnit() const -> const std::string & {
  return this->wrapped->unit();
}

[[nodiscard]] auto Property::GetDataType() const -> const DataType & { return this->data_type; }

auto Property::SetId(const std::string &id) -> void { this->wrapped->set_id(id); }

auto Property::SetName(const std::string &name) -> void { this->wrapped->set_name(name); }

auto Property::SetDescription(const std::string &description) -> void {
  this->wrapped->set_description(description);
}

auto Property::SetImportance(Importance importance) -> void {
  this->wrapped->set_importance(static_cast<ProtoImportance>(importance));
}

auto Property::SetValue(const Value &value) -> StatusCode { return this->value.CopyFrom(value); }

auto Property::SetValue(const std::string &value) -> StatusCode { return SetValue(Value(value)); }

auto Property::SetValue(const char *value) -> StatusCode { return SetValue(Value(value)); }

auto Property::SetValue(double value) -> StatusCode { return SetValue(Value(value)); }

auto Property::SetValue(const std::vector<std::string> &values) -> StatusCode {
  return SetValue(Value(values));
}

auto Property::SetValue(const std::vector<char *> &values) -> StatusCode {
  return SetValue(Value(values));
}

auto Property::SetValue(const std::vector<int64_t> &values) -> StatusCode {
  return SetValue(Value(values));
}

auto Property::SetValue(const std::vector<double> &values) -> StatusCode {
  return SetValue(Value(values));
}

auto Property::SetValue(const std::vector<bool> &values) -> StatusCode {
  return SetValue(Value(values));
}

auto Property::SetValue(const int64_t value) -> StatusCode { return SetValue(Value(value)); }

auto Property::SetValue(const bool value) -> StatusCode { return SetValue(Value(value)); }

auto Property::SetUnit(const std::string &unit) -> void { this->wrapped->set_unit(unit); }

auto Property::SetDataType(const DataType &new_data_type) -> StatusCode {
  return this->data_type.CopyFrom(new_data_type);
}

auto Property::SetDataType(const AtomicDataType new_data_type, bool list_type) -> StatusCode {
  return SetDataType(DataType(new_data_type, list_type));
}

auto Property::SetDataType(const std::string &new_data_type, bool list_type) -> StatusCode {
  return SetDataType(DataType(new_data_type, list_type));
}

auto Property::FixValue() -> void {
  FixValueImpl(this);
}

// Entity /////////////////////////////////////////////////////////////////////
[[nodiscard]] auto Entity::GetParents() const -> const Parents & { return parents; }

auto Entity::AppendParent(const Parent &parent) -> void { this->parents.Append(parent); }

auto Entity::RemoveParent(int index) -> void { this->parents.Remove(index); }

[[nodiscard]] auto Entity::GetProperties() const -> const Properties & { return properties; }

auto Entity::AppendProperty(const Property &property) -> void { this->properties.Append(property); }

auto Entity::RemoveProperty(int index) -> void { this->properties.Remove(index); }

auto Entity::CreateProtoEntity() -> ProtoEntity * {
  return Arena::CreateMessage<ProtoEntity>(get_arena());
}

Entity::Entity(IdResponse *id_response) : Entity() {
  this->wrapped->set_id(id_response->id());
  this->wrapped->mutable_version()->Swap(id_response->mutable_version());
  this->errors.wrapped->Swap(id_response->mutable_errors());
  this->warnings.wrapped->Swap(id_response->mutable_warnings());
  this->infos.wrapped->Swap(id_response->mutable_infos());
  FixValue();
}

Entity::Entity() : Entity(Entity::CreateProtoEntity()) {
  FixValue();
}

auto Entity::CreateMessagesField() -> RepeatedPtrField<ProtoMessage> * {
  return Arena::CreateMessage<RepeatedPtrField<ProtoMessage>>(get_arena());
}

auto Entity::SetId(const std::string &id) -> void { this->wrapped->set_id(id); }

auto Entity::SetVersionId(const std::string &id) -> void {
  this->wrapped->mutable_version()->set_id(id);
}

auto Entity::CopyTo(ProtoEntity *target) -> void { target->CopyFrom(*(this->wrapped)); }

auto Entity::SetRole(Role role) -> void { this->wrapped->set_role(static_cast<EntityRole>(role)); }

auto Entity::SetName(const std::string &name) -> void { this->wrapped->set_name(name); }

auto Entity::SetDescription(const std::string &description) -> void {
  this->wrapped->set_description(description);
}

auto Entity::SetValue(const Value &value) -> StatusCode {
  if (GetRole() != Role::PROPERTY) {
    return StatusCode::ENTITY_CANNOT_HAVE_A_VALUE;
  }
  return this->value.CopyFrom(value);
}

auto Entity::SetValue(const std::string &value) -> StatusCode { return SetValue(Value(value)); }

auto Entity::SetValue(const char *value) -> StatusCode { return SetValue(Value(value)); }

auto Entity::SetValue(double value) -> StatusCode { return SetValue(Value(value)); }

auto Entity::SetValue(const std::vector<std::string> &values) -> StatusCode {
  return SetValue(Value(values));
}

auto Entity::SetValue(const std::vector<char *> &values) -> StatusCode {
  return SetValue(Value(values));
}

auto Entity::SetValue(const std::vector<int64_t> &values) -> StatusCode {
  return SetValue(Value(values));
}

auto Entity::SetValue(const std::vector<double> &values) -> StatusCode {
  return SetValue(Value(values));
}

auto Entity::SetValue(const std::vector<bool> &values) -> StatusCode {
  return SetValue(Value(values));
}

auto Entity::SetValue(const int64_t value) -> StatusCode { return SetValue(Value(value)); }

auto Entity::SetValue(const bool value) -> StatusCode { return SetValue(Value(value)); }

auto Entity::SetUnit(const std::string &unit) -> void { this->wrapped->set_unit(unit); }

auto Entity::SetDataType(const DataType &new_data_type) -> StatusCode {
  if (GetRole() != Role::PROPERTY) {
    return StatusCode::ENTITY_CANNOT_HAVE_A_DATA_TYPE;
  }
  return this->data_type.CopyFrom(new_data_type);
}

auto Entity::SetDataType(const AtomicDataType new_data_type, bool list_type) -> StatusCode {
  return SetDataType(DataType(new_data_type, list_type));
}

auto Entity::SetDataType(const std::string &new_data_type, bool list_type) -> StatusCode {
  return SetDataType(DataType(new_data_type, list_type));
}

auto Entity::SetFilePath(const std::string &path) -> void {
  this->wrapped->mutable_file_descriptor()->set_path(path);
}

auto Entity::FixValue() -> void {
  FixValueImpl(this);
}

// Utility functions //////////////////////////////////////////////////////////

template<typename E>
auto FixValueImpl(E* ent) -> void {
  const auto &dtype = ent->GetDataType();
  const auto &value = ent->GetValue();
  auto new_value = Value();
  if (value.IsNull() || ! value.IsString()){ // Don't treat NULL and non-string values.
    return;
  }
  if (value.IsList()) { // Also don't treat empty or non-string lists.
    const auto &list = value.AsList();
    if (list.empty() || ! list[0].IsString()) {
      return;
    }
  }
  auto atype = AtomicDataType::UNSPECIFIED;
  if (dtype.IsList()) {                                     // List Datatype
    if (!value.IsList()) {
      throw caosdb::exceptions::Exception(StatusCode::OTHER_CLIENT_ERROR,
                                          "DataType is list, but Value is scalar.");
    }
    auto &list_type = dtype.AsList();
    atype = list_type.GetAtomicDataType();
    if (!list_type.IsListOfAtomic()  // References, strings etc. need no treatment.
        || atype == AtomicDataType::UNSPECIFIED
        || atype == AtomicDataType::TEXT
        || atype == AtomicDataType::DATETIME) {
      return;
    }
    if (atype == AtomicDataType::DOUBLE) {
      std::vector<double> data;
      for (auto &d: value.AsList()) {
        data.push_back(std::stod(d.AsString()));
      }
      new_value = Value(data) ;
    } else if (atype == AtomicDataType::INTEGER) {
      std::vector<long> data;
      for (auto &d: value.AsList()) {
        data.push_back(std::stol(d.AsString()));
      }
      new_value = Value(data) ;
    } else if (atype == AtomicDataType::BOOLEAN) {
      std::vector<bool> data;
      for (auto &d: value.AsList()) {
        auto bool_value = d.AsString();
        if (boost::to_upper_copy(bool_value) == "TRUE") {
          data.push_back(true);
        } else if (boost::to_upper_copy(bool_value) == "FALSE") {
          data.push_back(false);
        } else {
          throw caosdb::exceptions::Exception(StatusCode::OTHER_CLIENT_ERROR,
                                              "Boolean value is neither true nor false.");
        }
      }
      new_value = Value(data) ;
    } else {
      std::cout << "Unhandled datatype: " << ent->ToString() << std::endl;
      throw std::logic_error("Unhandled datatype");
    }
  } else {                                                  // Scalar Datatype
    if (value.IsList()) {
      throw caosdb::exceptions::Exception(StatusCode::OTHER_CLIENT_ERROR,
                                          "Value is list, but DataType is scalar.");
    }
    atype = dtype.AsAtomic();
    if (!dtype.IsAtomic()  // References, strings etc. need no treatment.
        || atype == AtomicDataType::UNSPECIFIED
        || atype == AtomicDataType::TEXT
        || atype == AtomicDataType::DATETIME) {
      return;
    }
    if (atype == AtomicDataType::DOUBLE) {
      new_value = Value(std::stod(value.AsString()));
    } else if (atype == AtomicDataType::INTEGER) {
      new_value = Value(std::stol(value.AsString()));
    } else if (atype == AtomicDataType::BOOLEAN) {
      auto bool_value = value.AsString();
      if (boost::to_upper_copy(bool_value) == "TRUE") {
        new_value = Value(true);
      } else if (boost::to_upper_copy(bool_value) == "FALSE") {
        new_value = Value(false);
      } else {
        throw caosdb::exceptions::Exception(StatusCode::OTHER_CLIENT_ERROR,
                                            "Boolean value is neither true nor false.");
      }
    } else {
      std::cout << "Unhandled datatype: " << ent->ToString() << std::endl;
      throw std::logic_error("Unhandled datatype");
    }
  }

  ent->SetValue(new_value);
}


} // namespace caosdb::entity
